//############################################################################
//
// EZFB ~ Linux Frame Buffer API ~
//
// by James Lehman
// james@akrobiz.com
//
// began: Feb. 2001
//
// EZFB is intended to give C programmers an easy-to-use library of functions
// for drawing points, lines and strings of characters into the Linux frame
// buffer. It also provides a means of displaying bitmap images on any portion
// of the screen and capturing the contents of any portion of the screen to
// a standard bitmap graphics file. All of this is embodied in a set of
// functions that are polymorphic with regard to the characteristics of the
// frame buffer and the bitmap files used. In other words, it makes no
// difference what the screen height or width is or what the color depth is,
// all of the function calls are the same and will have appropriate and
// predictable effects.
//
// This software is distributed in the hope that it will be useful, but
// without any warranty; without even the implied warranty of merchantability
// or fitness for a particular purpose. See the GNU General Public License
// for more details.
//
//############################################################################

#include "ezfb_device.h"
#include "ezfb_colors.h"
#include "ezfb_pixels.h"
#include "ezfb_bmp.h"

//############################################################################
// UTILITY ROUTINES
//############################################################################
int fix_rectangle(struct rectangle* area, int max_x, int max_y)
{
    int swap;
EZFB_DEBUGGING_FUNCTION_CALL

    if(0 > area->x1)            area->x1 = 0;
    if(0 > area->x2)            area->x2 = max_x;
    if(0 > area->y1)            area->y1 = 0;
    if(0 > area->y2)            area->y2 = max_y;
    if(area->x1 > area->x2)  {  swap = area->x2;  area->x2 = area->x1;  area->x1 = swap;  }
    if(area->y1 > area->y2)  {  swap = area->y2;  area->y2 = area->y1;  area->y1 = swap;  }
    if(max_x < area->x2)        area->x2 = max_x;
    if(max_y < area->y2)        area->y2 = max_y;
    area->height = area->y2 - area->y1 + 1;
    area->width  = area->x2 - area->x1 + 1;
EZFB_DEBUGGING_FUNCTION_RETURN(1)
}

//############################################################################
int dump_rectangle(struct rectangle* area, char* message)
{
EZFB_DEBUGGING_FUNCTION_CALL
#if EZFB_API_MESSAGES
    printf("\n             %s\n"         , message      );
    printf(  "             x1     = %d\n", area->x1     );
    printf(  "             y1     = %d\n", area->y1     );
    printf(  "             x2     = %d\n", area->x2     );
    printf(  "             y2     = %d\n", area->y2     );
    printf(  "             width  = %d\n", area->width  );
    printf(  "             height = %d\n", area->height );
    printf("\n"                          );fflush(stdout);
#endif
EZFB_DEBUGGING_FUNCTION_RETURN(1)
}

//############################################################################
int bmp_dump_header(struct bitmap_object* bmp_obj, char* file_name, char* message)
{
EZFB_DEBUGGING_FUNCTION_CALL
#if EZFB_API_MESSAGES
    printf("\n%s %s\n", message, file_name);
    printf(  "             B              = %c\n"  , bmp_obj->B              );
    printf(  "             M              = %c\n"  , bmp_obj->M              );
    printf(  "             file_size      = %d\n"  , bmp_obj->file_size      );
    printf(  "             reserved       = %d\n"  , bmp_obj->reserved       );
    printf(  "             jump           = %d\n"  , bmp_obj->jump           );
    printf(  "             to_end         = %d\n"  , bmp_obj->to_end         );
    printf(  "             xres           = %d\n"  , bmp_obj->xres           );
    printf(  "             yres           = %d\n"  , bmp_obj->yres           );
    printf(  "             planes         = %d\n"  , bmp_obj->planes         );
    printf(  "             bpp            = %d\n"  , bmp_obj->bpp            );
    printf(  "             comp           = %d\n"  , bmp_obj->comp           );
    printf(  "             image_size     = %d\n"  , bmp_obj->image_size     );
    printf(  "             xppm           = %d\n"  , bmp_obj->xppm           );
    printf(  "             yppm           = %d\n"  , bmp_obj->yppm           );
    printf(  "             c_used         = %d\n"  , bmp_obj->c_used         );
    printf(  "             c_important    = %d\n"  , bmp_obj->c_important    );
    printf(  "             fd             = %d\n"  , bmp_obj->fd             );
    printf(  "             bytes_per_line = %d\n"  , bmp_obj->bytes_per_line );
    printf(  "             palette_size   = %d\n"  , bmp_obj->palette_size   );
    printf(  "\n"                                             );fflush(stdout);
#endif
EZFB_DEBUGGING_FUNCTION_RETURN(1)
}

//############################################################################





















//############################################################################
// BITMAP AS MEMORY OBJECT ROUTINES
//############################################################################
int bmp_object_init(struct bitmap_object* bmp_obj, int xres, int yres, short bpp)
{
EZFB_DEBUGGING_FUNCTION_CALL
    bmp_obj->B          = 'B'                          ;
    bmp_obj->M          = 'M'                          ;
    bmp_obj->reserved   = 0                            ;
    bmp_obj->to_end     = SIZE_OF_BMP_FILE_HEADER - 14 ;
    bmp_obj->xres       = xres                         ;
    bmp_obj->yres       = yres                         ;
    bmp_obj->planes     = 1                            ;
    bmp_obj->comp       = 0                            ;
    bmp_obj->xppm       = 2835                         ;
    bmp_obj->yppm       = 2835                         ;
    bmp_obj->palette    = (u_int* )0                   ;
    bmp_obj->image      = (u_char*)0                   ;
    bmp_obj->data       = (u_char*)0                   ;
    switch(bpp) // bmp_obj->xres & bmp_obj->yres must already be set!
    {
        case  1: bmp_obj->bytes_per_line = (ceil((bmp_obj->xres * 0.125) / 4.0) * 4);
                 bmp_obj->bpp            = 1;
                 bmp_obj->palette_size   = 8;
                 bmp_obj->c_used         = 2;
                 bmp_obj->c_important    = 2;
                 break;
        // --------------------------------------------------------------- //
        case  2:
        case  4: bmp_obj->bytes_per_line = (ceil((bmp_obj->xres * 0.5) / 4.0) * 4);
                 bmp_obj->bpp            = 4;
                 bmp_obj->palette_size   = 64;
                 bmp_obj->c_used         = 16;
                 bmp_obj->c_important    = 16;
                 break;
        // --------------------------------------------------------------- //
        case  8: bmp_obj->bytes_per_line = (ceil( bmp_obj->xres / 4.0) * 4);
                 bmp_obj->bpp            = 8;
                 bmp_obj->palette_size   = 1024;
                 bmp_obj->c_used         = 256;
                 bmp_obj->c_important    = 256;
                 break;
        // --------------------------------------------------------------- //
        case 15:
        case 16:
        case 24:
        case 32: bmp_obj->bytes_per_line = (ceil((bmp_obj->xres * 3.0) / 4.0) * 4);
                 bmp_obj->bpp            = 24;
                 bmp_obj->palette_size   = 0;
                 bmp_obj->c_used         = 0;
                 bmp_obj->c_important    = 0;
                 break;
        // --------------------------------------------------------------- //
        default: bmp_obj->bytes_per_line = 0;
                 bmp_obj->bpp            = 0;
                 bmp_obj->palette_size   = 0;
                 bmp_obj->c_used         = 0;
                 bmp_obj->c_important    = 0;
                 break;
        // --------------------------------------------------------------- //
    }
    bmp_obj->image_size = bmp_obj->yres * bmp_obj->bytes_per_line        ;
    bmp_obj->jump       = SIZE_OF_BMP_FILE_HEADER + bmp_obj->palette_size ;
    bmp_obj->file_size  = bmp_obj->image_size + bmp_obj->jump            ;
EZFB_DEBUGGING_FUNCTION_RETURN(1)
}

//############################################################################
int bmp_malloc_object(struct bitmap_object* bmp_obj)
{
EZFB_DEBUGGING_FUNCTION_CALL
    if(bmp_obj->data) free(bmp_obj->data);
    bmp_obj->data    = malloc(bmp_obj->palette_size + bmp_obj->image_size);
    bmp_obj->palette = (u_int*)bmp_obj->data;
    bmp_obj->image   = bmp_obj->data + bmp_obj->palette_size;
EZFB_DEBUGGING_FUNCTION_RETURN(1)
}

//############################################################################
int bmp_object_put_pixel(struct bitmap_object* bmp_obj, int x, int y, u_int c)
{
    if(bmp_obj->data && (x < bmp_obj->xres) && (y < bmp_obj->yres))
    {
        switch(bmp_obj->bpp)
        {
            case  1: break;
            /* ----------------------------------------------------------------------------- */
            case  4: {
                         int i = (y * bmp_obj->xres + x) / 2;
                         bmp_obj->image[i] = (x & 0x01) // odd
                                             ? ((bmp_obj->image[i] & 0x0F) | ((c & 0x0F) << 4))
                                             : ((bmp_obj->image[i] & 0xF0) |  (c & 0x0F)      );
                     }
                     break;
            /* ----------------------------------------------------------------------------- */
            case  8: bmp_obj->image[y * bmp_obj->xres + x] = (u_char)c;
                     break;
            /* ----------------------------------------------------------------------------- */
            case 24: {
                         int i = (y * bmp_obj->xres + x) * 3;
                         bmp_obj->image[i++] = c & 0x000000ff;
                         bmp_obj->image[i++] = c & 0x0000ff00;
                         bmp_obj->image[i  ] = c & 0x00ff0000;
                     }
                     break;
            /* ----------------------------------------------------------------------------- */
            default: return 0;
        }
        return 1;
    }
    return 0;
}

//############################################################################
int bmp_object_get_pixel(struct bitmap_object* bmp_obj, int x, int y)
{
    if(bmp_obj->data && (x < bmp_obj->xres) && (y < bmp_obj->yres))
    {
        switch(bmp_obj->bpp)
        {
            case  1: return 0;
            /* ----------------------------------------------------------------------------- */
            case  4: {
                        int i = (y * bmp_obj->xres + x) / 2;
                        return (u_int)( (x & 0x01) // odd
                                          ? ((bmp_obj->image[i] & 0xF0) >> 4)
                                          :  (bmp_obj->image[i] & 0x0F)       );
                     }
            /* ----------------------------------------------------------------------------- */
            case  8: return (u_int)bmp_obj->image[y * bmp_obj->xres + x];
            /* ----------------------------------------------------------------------------- */
            case 24: {
                         int i = (y * bmp_obj->xres + x) * 3;
                         return   (bmp_obj->image[i++] << 16)
                                | (bmp_obj->image[i++] <<  8)
                                |  bmp_obj->image[i  ]        ;
                     }
            /* ----------------------------------------------------------------------------- */
            default: return -1;
        }
    }
    return -1;
}

//############################################################################
int bmp_object_line(struct bitmap_object* bmp_obj, int x1, int y1, int x2, int y2, u_int c)
{
    int ret = 0;
    int dx  = x2 - x1;
    int dy  = y2 - y1;
    int n, b;
    float m;

    if(0 == dx)
        for(n = y1; n <= y2; n += (dy > 0))
            ret = bmp_object_put_pixel(bmp_obj, x1, n, c);

    else if(0 == dy)
        for(n = x1; n <= x2; n += (dx > 0))
            ret = bmp_object_put_pixel(bmp_obj, n, y1, c);

    else if(abs(dx) >= abs(dy))
    {
        m = ((float)dy) / ((float)dx);
        b = (int)(y1 - m * x1);
        for(n = x1; n <= x2; n += (dx > 0))
            ret = bmp_object_put_pixel(bmp_obj, n, (int)(n * m + b), c);
    }
    else
    {
        m = ((float)dx) / ((float)dy);
        b = (int)(x1 - m * y1);
        for(n = y1; n <= y2; n += (dy > 0))
            ret = bmp_object_put_pixel(bmp_obj, (int)(n * m + b), n, c);
    }
    return ret;
}

//############################################################################
int bmp_object_free(struct bitmap_object* bmp_obj)
{
EZFB_DEBUGGING_FUNCTION_CALL
    if(bmp_obj->data) free(bmp_obj->data);
    bmp_obj->palette = (u_int*)
    bmp_obj->data    =
    bmp_obj->image   = (u_char*)0;
EZFB_DEBUGGING_FUNCTION_RETURN(1)
}

//############################################################################




















//############################################################################
// BITMAP OBJECT AND BMP FILE ROUTINES
//############################################################################
int bmp_malloc_object_from_file(struct bitmap_object* bmp_obj, char* file_name)
{
EZFB_DEBUGGING_FUNCTION_CALL_1
    ret &= bmp_open_for_read     (bmp_obj, file_name);
    ret &= bmp_read_file_header  (bmp_obj           );
    ret &= bmp_malloc_object     (bmp_obj           );
    ret &= bmp_read_object_image (bmp_obj           );
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int bmp_open_for_read(struct bitmap_object* bmp_obj, char* file_name)
{
EZFB_DEBUGGING_FUNCTION_CALL_1
    if(0 > (bmp_obj->fd = open(file_name, O_RDONLY))) // open the bitmap file
    {
#if EZFB_API_ERRORS
        perror("\nezfb ERROR: opening bmp file failed");
#endif
        ret = 0;
    }
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int bmp_read_file_header(struct bitmap_object* bmp_obj)
{
    int bytes_read = 0;
EZFB_DEBUGGING_FUNCTION_CALL_1
    // file descriptor must be open, valid and in bmp_obj->fd
    lseek(bmp_obj->fd, 0, SEEK_SET); // start at the top
    bytes_read += read(bmp_obj->fd, (void*)(&bmp_obj->B           ), sizeof(char) );
    bytes_read += read(bmp_obj->fd, (void*)(&bmp_obj->M           ), sizeof(char) );
    bytes_read += read(bmp_obj->fd, (void*)(&bmp_obj->file_size   ), sizeof(int)  );
    bytes_read += read(bmp_obj->fd, (void*)(&bmp_obj->reserved    ), sizeof(int)  );
    bytes_read += read(bmp_obj->fd, (void*)(&bmp_obj->jump        ), sizeof(int)  );
    bytes_read += read(bmp_obj->fd, (void*)(&bmp_obj->to_end      ), sizeof(int)  );
    bytes_read += read(bmp_obj->fd, (void*)(&bmp_obj->xres        ), sizeof(int)  );
    bytes_read += read(bmp_obj->fd, (void*)(&bmp_obj->yres        ), sizeof(int)  );
    bytes_read += read(bmp_obj->fd, (void*)(&bmp_obj->planes      ), sizeof(short));
    bytes_read += read(bmp_obj->fd, (void*)(&bmp_obj->bpp         ), sizeof(short));
    bytes_read += read(bmp_obj->fd, (void*)(&bmp_obj->comp        ), sizeof(int)  );
    bytes_read += read(bmp_obj->fd, (void*)(&bmp_obj->image_size  ), sizeof(int)  );
    bytes_read += read(bmp_obj->fd, (void*)(&bmp_obj->xppm        ), sizeof(int)  );
    bytes_read += read(bmp_obj->fd, (void*)(&bmp_obj->yppm        ), sizeof(int)  );
    bytes_read += read(bmp_obj->fd, (void*)(&bmp_obj->c_used      ), sizeof(int)  );
    bytes_read += read(bmp_obj->fd, (void*)(&bmp_obj->c_important ), sizeof(int)  );

    bmp_obj->bytes_per_line = (ceil((bmp_obj->bpp / 8.0) * bmp_obj->xres / 4.0) * 4);
    bmp_obj->palette_size   = bmp_obj->jump - SIZE_OF_BMP_FILE_HEADER;

    if(bytes_read != SIZE_OF_BMP_FILE_HEADER)
        ret = 0;

EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int bmp_read_object_palette(struct bitmap_object* bmp_obj)
{
EZFB_DEBUGGING_FUNCTION_CALL_0
    if(bmp_obj->data)
    {
        lseek(bmp_obj->fd, 14 + bmp_obj->to_end, SEEK_SET);
        ret = (bmp_obj->palette_size == read(bmp_obj->fd, (void*)(bmp_obj->palette), bmp_obj->palette_size));
    }
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int bmp_read_object_image(struct bitmap_object* bmp_obj)
{
EZFB_DEBUGGING_FUNCTION_CALL_0
    if(bmp_obj->data)
    {
        int bytes_read = 0,
            i;

        lseek(bmp_obj->fd, bmp_obj->jump, SEEK_SET);

        switch(bmp_obj->bpp)
        {
            case  1: for(i = 0; i < bmp_obj->yres; i++)
                     {
                         bytes_read += read(bmp_obj->fd, (void*)(bmp_obj->image), bmp_obj->xres / 8);
                         lseek(bmp_obj->fd, bmp_obj->bytes_per_line - bmp_obj->xres / 8, SEEK_CUR);
                     }
                     break;
            // ----------------------------------------------------------------- //
            case  4: for(i = 0; i < bmp_obj->yres; i++)
                     {
                         bytes_read += read(bmp_obj->fd, (void*)(bmp_obj->image), bmp_obj->xres / 2);
                         lseek(bmp_obj->fd, bmp_obj->bytes_per_line - bmp_obj->xres / 2, SEEK_CUR);
                     }
                     break;
            // ----------------------------------------------------------------- //
            case  8: for(i = 0; i < bmp_obj->yres; i++)
                     {
                         bytes_read += read(bmp_obj->fd, (void*)(bmp_obj->image), bmp_obj->xres);
                         lseek(bmp_obj->fd, bmp_obj->bytes_per_line - bmp_obj->xres, SEEK_CUR);
                     }
                     break;
            // ----------------------------------------------------------------- //
            case 24: for(i = 0; i < bmp_obj->yres; i++)
                     {
                         bytes_read += read(bmp_obj->fd, (void*)(bmp_obj->image), bmp_obj->xres * 3);
                         lseek(bmp_obj->fd, bmp_obj->bytes_per_line - bmp_obj->xres * 3, SEEK_CUR);
                     }
                     break;
            // ----------------------------------------------------------------- //
        }
        ret = (bmp_obj->image_size == bytes_read);
    }
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int bmp_open_for_write(struct bitmap_object* bmp_obj, char* file_name)
{
EZFB_DEBUGGING_FUNCTION_CALL_1
    if(0 > (bmp_obj->fd = open(file_name, (O_WRONLY | O_CREAT), (S_IRWXU | S_IRGRP | S_IROTH)))) // mode 744
    {
#if EZFB_API_ERRORS
        perror("\nezfb ERROR: opening bmp file failed");
#endif
        ret = 0;
    }
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int bmp_write_file_header(struct bitmap_object* bmp_obj)
{
    int bytes_written = 0;
EZFB_DEBUGGING_FUNCTION_CALL_1
    if(bmp_obj->fd)
    {
        bytes_written += write(bmp_obj->fd, (void*)(&bmp_obj->B           ), sizeof(char) );
        bytes_written += write(bmp_obj->fd, (void*)(&bmp_obj->M           ), sizeof(char) );
        bytes_written += write(bmp_obj->fd, (void*)(&bmp_obj->file_size   ), sizeof(int)  );
        bytes_written += write(bmp_obj->fd, (void*)(&bmp_obj->reserved    ), sizeof(int)  );
        bytes_written += write(bmp_obj->fd, (void*)(&bmp_obj->jump        ), sizeof(int)  );
        bytes_written += write(bmp_obj->fd, (void*)(&bmp_obj->to_end      ), sizeof(int)  );
        bytes_written += write(bmp_obj->fd, (void*)(&bmp_obj->xres        ), sizeof(int)  );
        bytes_written += write(bmp_obj->fd, (void*)(&bmp_obj->yres        ), sizeof(int)  );
        bytes_written += write(bmp_obj->fd, (void*)(&bmp_obj->planes      ), sizeof(short));
        bytes_written += write(bmp_obj->fd, (void*)(&bmp_obj->bpp         ), sizeof(short));
        bytes_written += write(bmp_obj->fd, (void*)(&bmp_obj->comp        ), sizeof(int)  );
        bytes_written += write(bmp_obj->fd, (void*)(&bmp_obj->image_size  ), sizeof(int)  );
        bytes_written += write(bmp_obj->fd, (void*)(&bmp_obj->xppm        ), sizeof(int)  );
        bytes_written += write(bmp_obj->fd, (void*)(&bmp_obj->yppm        ), sizeof(int)  );
        bytes_written += write(bmp_obj->fd, (void*)(&bmp_obj->c_used      ), sizeof(int)  );
        bytes_written += write(bmp_obj->fd, (void*)(&bmp_obj->c_important ), sizeof(int)  );
    }
    if(bytes_written != (14 + bmp_obj->to_end))
        ret = 0;
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int bmp_write_object_palette(struct bitmap_object* bmp_obj)
{
EZFB_DEBUGGING_FUNCTION_CALL_0
    ret = (bmp_obj->palette_size == write(bmp_obj->fd, (void*)(bmp_obj->palette), bmp_obj->palette_size));
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int bmp_write_object_image(struct bitmap_object* bmp_obj)
{
    int   i,
          bytes_written  = 0,
          zero           = 0;
EZFB_DEBUGGING_FUNCTION_CALL_0
    for(i = 0; i < bmp_obj->yres; i++)
    {
        switch(bmp_obj->bpp)
        {
            case  1: break;
            // ----------------------------------------------------------------- //
            case  4: {
                         int     j,
                                 local_x;
                         u_char  c;
                         for(local_x = 0; local_x < bmp_obj->xres / 2; local_x++)
                         {
                             j = i * bmp_obj->xres / 2 + local_x;
                             c = ((bmp_obj->image[j] & 0x0F) << 4) + ((bmp_obj->image[j] & 0xF0) >> 4); // nibble swap!
                             bytes_written += write(bmp_obj->fd, (void*)&c, sizeof(char));
                         }
                         bytes_written += write(bmp_obj->fd, (void*)&(zero), (bmp_obj->bytes_per_line - bmp_obj->xres / 2));
                     }
                     break;
            // ----------------------------------------------------------------- //
            case  8: bytes_written += write(bmp_obj->fd, (void*)&(bmp_obj->image[i * bmp_obj->xres]), bmp_obj->bytes_per_line);
                     bytes_written += write(bmp_obj->fd, (void*)&(zero), (bmp_obj->bytes_per_line - bmp_obj->xres));
                     break;
            // ----------------------------------------------------------------- //
            case 24: bytes_written += write(bmp_obj->fd, (void*)&(bmp_obj->image[i * bmp_obj->xres * 3]), bmp_obj->bytes_per_line);
                     bytes_written += write(bmp_obj->fd, (void*)&(zero), bmp_obj->bytes_per_line - bmp_obj->xres * 3);
                     break;
            // ----------------------------------------------------------------- //
        }
    }
    ret = (bytes_written == bmp_obj->image_size);
#if EZFB_API_MESSAGES
    printf("\n****** bytes_written = %d ******\n", bytes_written);fflush(stdout);
#endif
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int bmp_save_object_as(struct bitmap_object* bmp_obj, char* file_name)
{
EZFB_DEBUGGING_FUNCTION_CALL_1
    ret &= bmp_open_for_write       (bmp_obj, file_name);
    ret &= bmp_write_object_palette (bmp_obj);
    ret &= bmp_write_object_image   (bmp_obj);
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################




















//############################################################################
// FRAME BUFFER OUTPUT AS BMP FILE ROUTINES
//############################################################################
int bmp_write_palette_from_cmap(struct frame_buffer_device* fb, struct bitmap_object* bmp_obj)
{
    u_char c;
    int    i,
           bytes_written = 0;
EZFB_DEBUGGING_FUNCTION_CALL_1
    // file descriptor must be open, valid and in bmp_obj->fd
    if(bmp_obj->bpp != 24)
    {
        for(i = 0; i < bmp_obj->c_used; i++) // must be called after bmp_write_file_header(...
        {
            c = ((fb->colors.blue  [i] >> (fb->Var.blue.msb_right ? 0 : NUM_BITS_IN_CHAR)) & fb->colors.bit_mask_blue);
            bytes_written += write(bmp_obj->fd, (void*)(&c), sizeof(char));

            c = ((fb->colors.green [i] >> (fb->Var.green.msb_right ? 0 : NUM_BITS_IN_CHAR)) & fb->colors.bit_mask_green);
            bytes_written += write(bmp_obj->fd, (void*)(&c), sizeof(char));

            c = ((fb->colors.red   [i] >> (fb->Var.red.msb_right ? 0 : NUM_BITS_IN_CHAR)) & fb->colors.bit_mask_red);
            bytes_written += write(bmp_obj->fd, (void*)(&c), sizeof(char));

            c = (char)0;
            bytes_written += write(bmp_obj->fd, (void*)(&c), sizeof(char));
        }
        ret = (bytes_written == bmp_obj->palette_size);
    }
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int bmp_write_image_from_fb(struct frame_buffer_device* fb, struct bitmap_object* bmp_obj, struct rectangle* fb_area)
{
    int   i,
          bytes_written  = 0,
          zero           = 0;
EZFB_DEBUGGING_FUNCTION_CALL_0
    // must be called after bmp_write_file_header(... and bmp_write_palette_from_cmap(...
    for(i = fb_area->y2; i >= fb_area->y1; i--)
    {
        switch(fb->Var.bits_per_pixel)
        {
            case  1: break;
            // ----------------------------------------------------------------- //
            case  2: break;
            // ----------------------------------------------------------------- //
            case  4: {
                         int     j,
                                 local_x;
                         u_char  c;
                         for(local_x = fb_area->x1; local_x <= (fb_area->x2 / 2); local_x++)
                         {
                             j = i * fb->Var.xres_virtual / 2 + local_x;
                             c = ((fb->screen_uchars[j] & 0x0F) << 4) + ((fb->screen_uchars[j] & 0xF0) >> 4); // nibble swap!
                             bytes_written += write(bmp_obj->fd, (void*)&c, sizeof(char));
                         }
                         bytes_written += write(bmp_obj->fd, (void*)&(zero), (bmp_obj->bytes_per_line - fb_area->width / 2));
                     }
                     break;
            // ----------------------------------------------------------------- //
            default:
            case  8: bytes_written += write(bmp_obj->fd, (void*)&(fb->screen_uchars[i * fb->Var.xres_virtual + fb_area->x1]), fb_area->width);
                     bytes_written += write(bmp_obj->fd, (void*)&(zero), (bmp_obj->bytes_per_line - fb_area->width));
                     break;
            // ----------------------------------------------------------------- //
            case 15:
            case 16: {
                         int     j,
                                 local_x;
                         u_char  c;
                         for(local_x = fb_area->x1; local_x <= fb_area->x2; local_x++)
                         {
                             j = fb->screen_ushorts[i * fb->Var.xres_virtual + local_x];

                             c = get_b_from_rgb(fb, j);
                             bytes_written += write(bmp_obj->fd, (void*)&c, sizeof(char));

                             c = get_g_from_rgb(fb, j);
                             bytes_written += write(bmp_obj->fd, (void*)&c, sizeof(char));

                             c = get_r_from_rgb(fb, j);
                             bytes_written += write(bmp_obj->fd, (void*)&c, sizeof(char));
                         }
                         bytes_written += write(bmp_obj->fd, (void*)&(zero), (bmp_obj->bytes_per_line - (fb_area->width * 3)));
                     }
                     break;
            // ----------------------------------------------------------------- //
            case 24: bytes_written += write(bmp_obj->fd, (void*)&(fb->screen_uchars[(i * fb->Var.xres_virtual + fb_area->x1) * 3]), fb_area->width * 3);
                     bytes_written += write(bmp_obj->fd, (void*)&(zero), (bmp_obj->bytes_per_line - (fb_area->width * 3)));
                     break;
            // ----------------------------------------------------------------- //
            case 32: {
                         int     j,
                                 local_x;
                         u_char  c;
                         for(local_x = fb_area->x1; local_x <= fb_area->x2; local_x++)
                         {
                             j = fb->screen_uints[i * fb->Var.xres_virtual + local_x];

                             c = (char)get_b_from_rgb(fb, j);
                             bytes_written += write(bmp_obj->fd, (void*)&c, sizeof(char));

                             c = (char)get_g_from_rgb(fb, j);
                             bytes_written += write(bmp_obj->fd, (void*)&c, sizeof(char));

                             c = (char)get_r_from_rgb(fb, j);
                             bytes_written += write(bmp_obj->fd, (void*)&c, sizeof(char));
                         }
                         bytes_written += write(bmp_obj->fd, (void*)&(zero), (bmp_obj->bytes_per_line - (fb_area->width * 3)));
                     }
                     break;
            // ----------------------------------------------------------------- //
        }
    }
    ret = (bytes_written == bmp_obj->image_size);
#if EZFB_API_MESSAGES
    printf("\n****** bytes_written = %d ******\n", bytes_written);fflush(stdout);
#endif
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int bmp_save_fb_as(struct frame_buffer_device* fb, char* file_name)
{
    struct rectangle        fb_area = { -1, -1, -1, -1, 0, 0 };
EZFB_DEBUGGING_FUNCTION_CALL_0
    ret = bmp_save_fb_area_as(fb, file_name, &fb_area);
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int bmp_save_fb_area_as(struct frame_buffer_device* fb, char* file_name, struct rectangle* fb_area)
{
    struct bitmap_object  bmp_obj;
EZFB_DEBUGGING_FUNCTION_CALL_0
    if(!fb->was_grabbed)
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "\nezfb ERROR: FB NOT GRABBED!");fflush(stderr);
#endif
    }
    else if(bmp_open_for_write(&bmp_obj, file_name))
    {
        fix_rectangle(fb_area, fb->max_x, fb->max_y);
        dump_rectangle(fb_area, "fb_area");
        ret  = bmp_object_init       (&bmp_obj, fb->Var.xres, fb->Var.yres, fb->Var.bits_per_pixel);
        ret &= bmp_write_file_header (&bmp_obj);
        bmp_dump_header              (&bmp_obj, file_name, "SAVING .....");
        if(bmp_obj.bpp != 24)
            ret &= bmp_write_palette_from_cmap(fb, &bmp_obj);
        ret &= bmp_write_image_from_fb(fb, &bmp_obj, fb_area);
        close(bmp_obj.fd);
    }
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################




















//############################################################################
// FRAME BUFFER AND BMP MEMORY OBJECT ROUTINES
//############################################################################
//int bmp_object_into_fb(struct frame_buffer_device* fb, struct bitmap_object* bmp_obj)
//############################################################################
//int bmp_malloc_object_from_fb(struct frame_buffer_device* fb, struct bitmap_object* bmp_obj)
//############################################################################
//int bmp_object_palette_from_cmap(struct frame_buffer_device* fb, struct bitmap_object* bmp_obj)
//############################################################################
//int bmp_cmap_from_object_palette(struct frame_buffer_device* fb, struct bitmap_object* bmp_obj)
//############################################################################




















//############################################################################
// BMP FILE AS INPUT TO FRAMEBUFFER ROUTINES
//############################################################################
int bmp_read_palette_into_cmap(struct frame_buffer_device* fb, struct bitmap_object* bmp_obj)
{
    int bytes_read = 0;
EZFB_DEBUGGING_FUNCTION_CALL_1
    // file descriptor must be open, valid and in bmp_obj->fd
    if(bmp_obj->bpp != 24)
    {
        char r, g, b, a;
        int  i;
        lseek(bmp_obj->fd, 14 + bmp_obj->to_end, SEEK_SET); // jump to end of header / beginning of palette
        for(i = 0; i < (1 << bmp_obj->bpp); i++)
        {
            bytes_read += read(bmp_obj->fd, (void*)(&b), sizeof(char));
            bytes_read += read(bmp_obj->fd, (void*)(&g), sizeof(char));
            bytes_read += read(bmp_obj->fd, (void*)(&r), sizeof(char));
            bytes_read += read(bmp_obj->fd, (void*)(&a), sizeof(char));
            set_fb_cmap_index (fb, i, r << NUM_BITS_IN_CHAR,
                                      g << NUM_BITS_IN_CHAR,
                                      b << NUM_BITS_IN_CHAR);
        }
        if(bytes_read == ((1 << bmp_obj->bpp) * 4))
        {
            if(fb->Var.bits_per_pixel <= 8)
                ret &= set_fb_cmap(fb);
        }
        else
            ret = 0;
    }
    else if(fb->Var.bits_per_pixel <= 8) // bmp_obj->bpp must be 24
        ret &= set_fb_cmap_332(fb);

    find_black_and_white(fb);
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int bmp_read_image_into_fb(struct frame_buffer_device* fb, struct bitmap_object* bmp_obj, struct rectangle* fb_area, struct rectangle* bmp_area)
{
    int   i,
          junk_before,
          junk_after,
          bytes_read  = 0;
EZFB_DEBUGGING_FUNCTION_CALL_0
    fix_rectangle(fb_area , fb->max_x                    , fb->max_y                     );
    fix_rectangle(bmp_area, bmp_obj->xres - 1            , bmp_obj->yres - 1             );
    fix_rectangle(bmp_area, bmp_area->x1 + fb_area->width, bmp_area->y1 + fb_area->height);
    junk_before = bmp_area->x1 * (bmp_obj->bpp / 8.0);
    junk_after  = bmp_obj->bytes_per_line - ((bmp_area->x1 + bmp_area->width) * (bmp_obj->bpp / 8.0));
    lseek(bmp_obj->fd, bmp_obj->jump, SEEK_SET); // jump over the header & palette to the image
    lseek(bmp_obj->fd, (bmp_obj->yres - bmp_area->y2 - 1) * bmp_obj->bytes_per_line, SEEK_CUR);
    for(i = (fb_area->y1 + bmp_area->height - 1); i >= fb_area->y1; i--)
    {
        switch(fb->Var.bits_per_pixel)
        {
            case  1:
            case  2: break;
            // ----------------------------------------------------------------- //
            case  4: switch(bmp_obj->bpp)
                     {
                        case  1: break;
                        // ----------------------------------------------------- //
                        case  4: {
                                     int     local_x;
                                     u_char  c;
                                     lseek(bmp_obj->fd, junk_before, SEEK_CUR);
                                     for(local_x = 0; local_x < bmp_area->width; local_x += 2)
                                     {
                                         bytes_read += read(bmp_obj->fd, (void*)&c, sizeof(char));
                                         put_pixel(fb, fb_area->x1 + local_x    , i, (c & 0xF0) >> 4);
                                         put_pixel(fb, fb_area->x1 + local_x + 1, i,  c & 0x0F      );
                                     }
                                     lseek(bmp_obj->fd, junk_after, SEEK_CUR);
                                 }
                                 break;
                        // ----------------------------------------------------- //
                        case  8: break;
                        // ----------------------------------------------------- //
                        case 24: {
                                     int    local_x;
                                     u_char r, g, b;
                                     lseek(bmp_obj->fd, junk_before, SEEK_CUR);
                                     for(local_x = 0; local_x < bmp_area->width * 3; local_x += 3)
                                     {
                                         bytes_read += read(bmp_obj->fd, (void*)&b, sizeof(char));
                                         bytes_read += read(bmp_obj->fd, (void*)&g, sizeof(char));
                                         bytes_read += read(bmp_obj->fd, (void*)&r, sizeof(char));
                                         put_pixel(fb, fb_area->x1 + local_x / 3, i, (r & 0xe0) | ((g & 0xe0) >> 3) | ((b & 0xc0) >> 6) / 17);
                                     }
                                     lseek(bmp_obj->fd, junk_after, SEEK_CUR);
                                 }
                                 break;
                        // ----------------------------------------------------- //
                     } // END switch(bmp_obj->bpp)
                     break;
            // ----------------------------------------------------------------- //
            default:
            case  8: switch(bmp_obj->bpp)
                     {
                        case  1: break;
                        // ----------------------------------------------------- //
                        case  4: {
                                     int     local_x;
                                     u_char  c;
                                     lseek(bmp_obj->fd, junk_before, SEEK_CUR);
                                     for(local_x = 0; local_x < bmp_area->width; local_x += 2)
                                     {
                                         bytes_read += read(bmp_obj->fd, (void*)&c, sizeof(char));
                                         put_pixel(fb, fb_area->x1 + local_x    , i, (c & 0xF0) >> 4);
                                         put_pixel(fb, fb_area->x1 + local_x + 1, i,  c & 0x0F      );
                                     }
                                     lseek(bmp_obj->fd, junk_after, SEEK_CUR);
                                 }
                                 break;
                        // ----------------------------------------------------- //
                        case  8: lseek(bmp_obj->fd, junk_before, SEEK_CUR);
                                 bytes_read += read(bmp_obj->fd, (void*)&(fb->screen_uchars[i * fb->Var.xres_virtual + fb_area->x1]), bmp_area->width);
                                 lseek(bmp_obj->fd, junk_after, SEEK_CUR);
                                 break;
                        // ----------------------------------------------------- //
                        case 24: {
                                     int    local_x;
                                     u_char r, g, b;
                                     lseek(bmp_obj->fd, junk_before, SEEK_CUR);
                                     for(local_x = 0; local_x < bmp_area->width * 3; local_x += 3)
                                     {
                                         bytes_read += read(bmp_obj->fd, (void*)&b, sizeof(char));
                                         bytes_read += read(bmp_obj->fd, (void*)&g, sizeof(char));
                                         bytes_read += read(bmp_obj->fd, (void*)&r, sizeof(char));
                                         put_pixel(fb, fb_area->x1 + local_x / 3, i, (r & 0xe0) | ((g & 0xe0) >> 3) | ((b & 0xc0) >> 6));
                                     }
                                     lseek(bmp_obj->fd, junk_after, SEEK_CUR);
                                 }
                                 break;
                        // ----------------------------------------------------- //
                     } // END switch(bmp_obj->bpp)
                     break;
            // ----------------------------------------------------------------- //
            case 15:
            case 16:
            case 24:
            case 32: switch(bmp_obj->bpp)
                     {
                        case  1: break;
                        // ----------------------------------------------------- //
                        case  4: {
                                     int    local_x;
                                     u_char c;
                                     lseek(bmp_obj->fd, junk_before, SEEK_CUR);
                                     for(local_x = 0; local_x < bmp_area->width; local_x += 2)
                                     {
                                         bytes_read += read(bmp_obj->fd, (void*)&c, sizeof(char));
                                         // ------------------------------------ //
                                         put_pixel(fb, fb_area->x1 + local_x,
                                                       i,
                                                       make_rgb( fb,
                                                                 fb->Var.red.msb_right
                                                                  ?  fb->colors.red   [(c & 0xF0) >> 4]
                                                                  : (fb->colors.red   [(c & 0xF0) >> 4] >> NUM_BITS_IN_CHAR),
                                                                 fb->Var.green.msb_right
                                                                  ?  fb->colors.green [(c & 0xF0) >> 4]
                                                                  : (fb->colors.green [(c & 0xF0) >> 4] >> NUM_BITS_IN_CHAR),
                                                                 fb->Var.blue.msb_right
                                                                  ?  fb->colors.blue  [(c & 0xF0) >> 4]
                                                                  : (fb->colors.blue  [(c & 0xF0) >> 4] >> NUM_BITS_IN_CHAR)
                                                               )
                                                  );
                                         // ------------------------------------ //
                                         put_pixel(fb, fb_area->x1 + local_x + 1,
                                                       i,
                                                       make_rgb( fb,
                                                                 fb->Var.red.msb_right
                                                                  ?  fb->colors.red   [c & 0x0F]
                                                                  : (fb->colors.red   [c & 0x0F] >> NUM_BITS_IN_CHAR),
                                                                 fb->Var.green.msb_right
                                                                  ?  fb->colors.green [c & 0x0F]
                                                                  : (fb->colors.green [c & 0x0F] >> NUM_BITS_IN_CHAR),
                                                                 fb->Var.blue.msb_right
                                                                  ?  fb->colors.blue  [c & 0x0F]
                                                                  : (fb->colors.blue  [c & 0x0F] >> NUM_BITS_IN_CHAR)
                                                               )
                                                  );
                                         // ------------------------------------ //
                                     }
                                     lseek(bmp_obj->fd, junk_after, SEEK_CUR);
                                 }
                                 break;
                        // ----------------------------------------------------- //
                        case  8: {
                                     int    local_x;
                                     u_char c;
                                     lseek(bmp_obj->fd, junk_before, SEEK_CUR);
                                     for(local_x = 0; local_x < bmp_area->width; local_x++)
                                     {
                                         bytes_read += read(bmp_obj->fd, (void*)&c, sizeof(char));
                                         put_pixel(fb, fb_area->x1 + local_x,
                                                       i,
                                                       make_rgb( fb,
                                                                 fb->Var.red.msb_right
                                                                  ?  fb->colors.red   [c]
                                                                  : (fb->colors.red   [c] >> NUM_BITS_IN_CHAR),
                                                                 fb->Var.green.msb_right
                                                                  ?  fb->colors.green [c]
                                                                  : (fb->colors.green [c] >> NUM_BITS_IN_CHAR),
                                                                 fb->Var.blue.msb_right
                                                                  ?  fb->colors.blue  [c]
                                                                  : (fb->colors.blue  [c] >> NUM_BITS_IN_CHAR)
                                                               )
                                                  );
                                     }
                                     lseek(bmp_obj->fd, junk_after, SEEK_CUR);
                                 }
                                 break;
                        // ----------------------------------------------------- //
                        case 24: {
                                     int    local_x;
                                     u_char r, g, b;
                                     lseek(bmp_obj->fd, junk_before, SEEK_CUR);
                                     for(local_x = 0; local_x < bmp_area->width; local_x++)
                                     {
                                         bytes_read += read(bmp_obj->fd, (void*)&b, sizeof(char));
                                         bytes_read += read(bmp_obj->fd, (void*)&g, sizeof(char));
                                         bytes_read += read(bmp_obj->fd, (void*)&r, sizeof(char));
                                         put_pixel(fb, fb_area->x1 + local_x, i, make_rgb(fb, r, g, b));
                                     }
                                     lseek(bmp_obj->fd, junk_after, SEEK_CUR);
                                 }
                                 break;
                        // ----------------------------------------------------- //
                     } // END switch(bmp_obj->bpp)
                     break;
            // ----------------------------------------------------------------- //
        } // END switch(fb->Var.bits_per_pixel)
    } // END for(i = (fb_area->y1 + bmp_area->height); i > fb_area->y1; i--)
    ret = (bytes_read == bmp_obj->image_size);
#if EZFB_API_MESSAGES
    printf("\n****** bytes_read = %d ******\n", bytes_read);fflush(stdout);
#endif
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int bmp_open_into_fb(struct frame_buffer_device* fb, char* file_name)
{
    struct rectangle fb_area  = { -1, -1, -1, -1 , 0, 0 };
    struct rectangle bmp_area = { -1, -1, -1, -1 , 0, 0 };
EZFB_DEBUGGING_FUNCTION_CALL_0
    ret = bmp_open_area_into_fb_at(fb, file_name, &fb_area, &bmp_area);
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int bmp_open_into_fb_at(struct frame_buffer_device* fb, char* file_name, struct rectangle* fb_area)
{
    struct rectangle bmp_area = { -1, -1, -1, -1 , 0, 0 };
EZFB_DEBUGGING_FUNCTION_CALL_0
    ret = bmp_open_area_into_fb_at(fb, file_name, fb_area, &bmp_area);
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int bmp_open_area_into_fb_at(struct frame_buffer_device* fb, char* file_name, struct rectangle* fb_area, struct rectangle* bmp_area)
{
    struct bitmap_object  bmp_obj;
EZFB_DEBUGGING_FUNCTION_CALL_0
    if(!fb->was_grabbed)
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "\nezfb ERROR: FB NOT GRABBED!");fflush(stderr);
#endif
    }
    else if(bmp_open_for_read(&bmp_obj, file_name)) // open the bitmap file
    {
        ret = bmp_read_file_header(&bmp_obj);
        bmp_dump_header(&bmp_obj, file_name, "OPENING ....");
        ret &= bmp_read_palette_into_cmap(fb, &bmp_obj);
        ret &= bmp_read_image_into_fb(fb, &bmp_obj, fb_area, bmp_area);
        close(bmp_obj.fd);
    }
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}



//############################################################################
int bmp_open_into_fb_center(struct frame_buffer_device* fb, char* file_name)
{
    struct bitmap_object  bmp_obj;
    struct rectangle fb_area  = { -1, -1, -1, -1 , 0, 0 };
    struct rectangle bmp_area = { -1, -1, -1, -1 , 0, 0 };
EZFB_DEBUGGING_FUNCTION_CALL_0
    if(!fb->was_grabbed)
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "\nezfb ERROR: FB NOT GRABBED!");fflush(stderr);
#endif
    }
    else if(bmp_open_for_read(&bmp_obj, file_name)) // open the bitmap file
    {
        ret = bmp_read_file_header(&bmp_obj);
        bmp_dump_header(&bmp_obj, file_name, "OPENING ....");
        ret &= bmp_read_palette_into_cmap(fb, &bmp_obj);
        if(bmp_obj.xres < fb->Var.xres)
            fb_area.x1 = (fb->max_x - bmp_obj.xres) / 2;
        if(bmp_obj.yres < fb->Var.yres)
            fb_area.y1 = (fb->max_y - bmp_obj.yres) / 2;
        dump_rectangle(&fb_area, "fb_area");
        ret &= bmp_read_image_into_fb(fb, &bmp_obj, &fb_area, &bmp_area);
            close(bmp_obj.fd);
    }
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int bmp_open_palette_into_fb_cmap(struct frame_buffer_device* fb, char* file_name)
{
    struct bitmap_object  bmp_obj;
EZFB_DEBUGGING_FUNCTION_CALL_0
    if(!fb->was_grabbed)
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "\nezfb ERROR: FB NOT GRABBED!");fflush(stderr);
#endif
    }
    else if(bmp_open_for_read(&bmp_obj, file_name)) // open the bitmap file
    {
        ret = bmp_read_file_header(&bmp_obj);
        bmp_dump_header(&bmp_obj, file_name, "OPENING .... ");
        ret &= bmp_read_palette_into_cmap(fb, &bmp_obj);
        close(bmp_obj.fd);
    }
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################

