//############################################################################
//
// EZFB ~ Linux Frame Buffer API ~
//
// by James Lehman
// james@akrobiz.com
//
// began: Feb. 2001
//
// EZFB is intended to give C programmers an easy-to-use library of functions
// for drawing points, lines and strings of characters into the Linux frame
// buffer. It also provides a means of displaying bitmap images on any portion
// of the screen and capturing the contents of any portion of the screen to
// a standard bitmap graphics file. All of this is embodied in a set of
// functions that are polymorphic with regard to the characteristics of the
// frame buffer and the bitmap files used. In other words, it makes no
// difference what the screen height or width is or what the color depth is,
// all of the function calls are the same and will have appropriate and
// predictable effects.
//
// This software is distributed in the hope that it will be useful, but
// without any warranty; without even the implied warranty of merchantability
// or fitness for a particular purpose. See the GNU General Public License
// for more details.
//
//############################################################################

#include "ezfb_device.h"
#include "ezfb_colors.h"

//############################################################################
int get_fb_cmap(struct frame_buffer_device* fb)
{
EZFB_DEBUGGING_FUNCTION_CALL_1
    if(0 > ioctl(fb->fd, FBIOGETCMAP, &fb->colors.fbcmap))
    {
#if EZFB_API_ERRORS
        perror("\nezfb ERROR: FBIOGETCMAP failed");
#endif
        ret = 0;
    }
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int set_fb_cmap(struct frame_buffer_device* fb)
{
EZFB_DEBUGGING_FUNCTION_CALL_1
    if(0 > ioctl(fb->fd, FBIOPUTCMAP, &fb->colors.fbcmap))
    {
#if EZFB_API_ERRORS
        perror("\nezfb ERROR: FBIOPUTCMAP failed");
#endif
        ret = 0;
    }
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int get_color_factors(struct frame_buffer_device* fb)
{
EZFB_DEBUGGING_FUNCTION_CALL_1
    fb->colors.shortest_color = 17; // one higher than possible
    if(fb->colors.shortest_color > fb->Var.red.length)
        fb->colors.shortest_color = fb->Var.red.length;
    if(fb->colors.shortest_color > fb->Var.green.length)
        fb->colors.shortest_color = fb->Var.green.length;
    if(fb->colors.shortest_color > fb->Var.blue.length)
        fb->colors.shortest_color = fb->Var.blue.length;
    // masks a u_char msb left.
    fb->colors.bit_mask_red   = ((1 << fb->Var.red.length  ) - 1) << (NUM_BITS_IN_CHAR - fb->Var.red.length  );
    fb->colors.bit_mask_green = ((1 << fb->Var.green.length) - 1) << (NUM_BITS_IN_CHAR - fb->Var.green.length);
    fb->colors.bit_mask_blue  = ((1 << fb->Var.blue.length ) - 1) << (NUM_BITS_IN_CHAR - fb->Var.blue.length );
    fb->colors.offset_plus_length_red   = fb->Var.red.offset   + fb->Var.red.length  ;
    fb->colors.offset_plus_length_green = fb->Var.green.offset + fb->Var.green.length;
    fb->colors.offset_plus_length_blue  = fb->Var.blue.offset  + fb->Var.blue.length ;
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int init_fb_cmap(struct frame_buffer_device* fb, int save_old)
{
EZFB_DEBUGGING_FUNCTION_CALL_1
    if(!fb->colors.was_grabbed)
    {
        fb->colors.fbcmap.start  = 0   ;
        fb->colors.fbcmap.transp = NULL;
        fb->colors.fbcmap.len    = (fb->Var.bits_per_pixel <= 8)
                                     ? (0x01 << fb->Var.bits_per_pixel)
                                     : 0;
        if(save_old)
        {
            fb->colors.fbcmap.red    = fb->colors.red_was   ;
            fb->colors.fbcmap.green  = fb->colors.green_was ;
            fb->colors.fbcmap.blue   = fb->colors.blue_was  ;
            ret &= get_fb_cmap         (fb);
            if(ret)
                fb->colors.old_was_saved = 1;
            fb->colors.fbcmap.red    = fb->colors.red       ; // the ones we're really going to use
            fb->colors.fbcmap.green  = fb->colors.green     ;
            fb->colors.fbcmap.blue   = fb->colors.blue      ;
            if(ret)
                fb->colors.was_grabbed = 1;
            ret &= get_color_factors   (fb);
            ret &= clear_fb_cmap       (fb);
            ret &= set_fb_cmap_utility (fb);
        }
        else
        {
            fb->colors.fbcmap.red    = fb->colors.red       ;
            fb->colors.fbcmap.green  = fb->colors.green     ;
            fb->colors.fbcmap.blue   = fb->colors.blue      ;
            ret &= get_fb_cmap         (fb);
            ret &= get_color_factors   (fb);
            if(ret)
            {
                fb->colors.was_grabbed   = 1;
                fb->colors.old_was_saved = 0;
            }
        }
    }
    else
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "\nezfb ERROR: init_fb_cmap(...) CALLED MORE THAN ONCE!");fflush(stderr);
#endif
        ret = 0;
    }
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int clear_fb_cmap(struct frame_buffer_device* fb)
{
    int i;
EZFB_DEBUGGING_FUNCTION_CALL_0
    for(i = 0; i < 256; i++)
        fb->colors.red[i] = fb->colors.green[i] = fb->colors.blue[i] = 0;
    ret = set_fb_cmap(fb);
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int release_fb_cmap(struct frame_buffer_device* fb)
{
EZFB_DEBUGGING_FUNCTION_CALL_1
    if(fb->colors.was_grabbed)
    {
        if(fb->colors.old_was_saved)
        {
            fb->colors.fbcmap.red    = fb->colors.red_was   ;
            fb->colors.fbcmap.green  = fb->colors.green_was ;
            fb->colors.fbcmap.blue   = fb->colors.blue_was  ;
            ret &= set_fb_cmap(fb); // put back the old color map
        }
        fb->colors.was_grabbed = 0;
    }
    else
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "\nezfb ERROR: FB CMAP NOT GRABBED!");fflush(stderr);
#endif
        ret = 0;
    }
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int find_black_and_white(struct frame_buffer_device* fb)
{
EZFB_DEBUGGING_FUNCTION_CALL
    switch(fb->Var.bits_per_pixel)
    {
        case  1: fb->colors.black = 0;
                 fb->colors.white = 1;
                 break;
        // ----------------------------------------------------------------------------- //
        case  2:
        case  4:
        case  8: {
                     int   i;
                     fb->colors.black = fb->colors.white = 0;
                     for(i = 0; i < fb->colors.fbcmap.len; i++)
                     {
                         if(   (    (u_int)fb->colors.red   [fb->colors.white]
                                  + (u_int)fb->colors.green [fb->colors.white]
                                  + (u_int)fb->colors.blue  [fb->colors.white]
                               )
                             < (    (u_int)fb->colors.red   [i]
                                  + (u_int)fb->colors.green [i]
                                  + (u_int)fb->colors.blue  [i]
                               )
                           )
                             fb->colors.white = i;
                           // ---------------------------------------------------------- //

                         if(   (    (u_int)fb->colors.red   [fb->colors.black]
                                  + (u_int)fb->colors.green [fb->colors.black]
                                  + (u_int)fb->colors.blue  [fb->colors.black]
                               )
                             > (    (u_int)fb->colors.red   [i]
                                  + (u_int)fb->colors.green [i]
                                  + (u_int)fb->colors.blue  [i]
                               )
                           )
                             fb->colors.black = i;
                           // ---------------------------------------------------------- //
                     }
                 }
                 break;
        // ----------------------------------------------------------------------------- //
        case 15:
        case 16:
        case 24:
        case 32:
        default: fb->colors.black = 0x00000000;
                 fb->colors.white = 0x00ffffff;
                 break;
    } // END switch(fb->Var.bits_per_pixel)
EZFB_DEBUGGING_FUNCTION_RETURN(fb->colors.was_grabbed)
}

//############################################################################
u_int make_rgb(struct frame_buffer_device* fb, u_short r, u_short g, u_short b)
{
    switch(fb->Var.bits_per_pixel)
    {
        default:
        case  1:
        case  2:
        case  4: return 0;
        // ------------------------------------------------------------------- //
        case  8: return (    (r & 0xe0)
                          | ((g & 0xe0) >> 3)
                          | ((b & 0xc0) >> 6)
                        );  // based on set_fb_cmap_332();
        // ------------------------------------------------------------------- //
        case 15:
        case 16: return ((   ((r & fb->colors.bit_mask_red  ) << fb->colors.offset_plus_length_red  )
                           | ((g & fb->colors.bit_mask_green) << fb->colors.offset_plus_length_green)
                           | ((b & fb->colors.bit_mask_blue ) << fb->colors.offset_plus_length_blue )  ) >> NUM_BITS_IN_CHAR);
        // ------------------------------------------------------------------- //
        case 24:
        case 32: return   ((r & fb->colors.bit_mask_red  ) << fb->Var.red.offset  )
                        | ((g & fb->colors.bit_mask_green) << fb->Var.green.offset)
                        | ((b & fb->colors.bit_mask_blue ) << fb->Var.blue.offset );
    }
}

//############################################################################
u_short get_r_from_rgb(struct frame_buffer_device* fb, u_int rgb)
{
    switch(fb->Var.bits_per_pixel)
    {
        default:
        case  1:
        case  2:
        case  4: return 0;
        // ------------------------------------------------------------------- //
        case  8: return (rgb & 0xe0); // based on set_fb_cmap_332();
        // ------------------------------------------------------------------- //
        case 15:     // returns a scaled char ie: fb->Var.red.length bits to 8 bits
        case 16: {   // by adding the top msb's to the otherwise zeros at the low end
                     u_char c = ((((rgb) << NUM_BITS_IN_CHAR) >> fb->colors.offset_plus_length_red) & fb->colors.bit_mask_red);
                     return c |= ((c & ~((1 << fb->Var.red.length) - 1)) >> fb->Var.red.length);
                 }   // off is off and full on is 255
        // ------------------------------------------------------------------- //
        case 24:
        case 32: return (rgb >> fb->Var.red.offset) & fb->colors.bit_mask_red;
    }
}

//############################################################################
u_short get_g_from_rgb(struct frame_buffer_device* fb, u_int rgb)
{
    switch(fb->Var.bits_per_pixel)
    {
        default:
        case  1:
        case  2:
        case  4: return 0;
        // ------------------------------------------------------------------- //
        case  8: return ((rgb << 3) & 0xe0); // based on set_fb_cmap_332();
        // ------------------------------------------------------------------- //
        case 15:
        case 16: {
                     u_char c = ((((rgb) << NUM_BITS_IN_CHAR) >> fb->colors.offset_plus_length_green) & fb->colors.bit_mask_green);
                     return c |= ((c & ~((1 << fb->Var.green.length) - 1)) >> fb->Var.green.length);
                 }
        // ------------------------------------------------------------------- //
        case 24:
        case 32: return (rgb >> fb->Var.green.offset) & fb->colors.bit_mask_green;
    }
}

//############################################################################
u_short get_b_from_rgb(struct frame_buffer_device* fb, u_int rgb)
{
    switch(fb->Var.bits_per_pixel)
    {
        default:
        case  1:
        case  2:
        case  4: return 0;
        // ------------------------------------------------------------------- //
        case  8: return ((rgb << 6) & 0xc0); // based on set_fb_cmap_332();
        // ------------------------------------------------------------------- //
        case 15:
        case 16: {
                     u_char c = ((((rgb) << NUM_BITS_IN_CHAR) >> fb->colors.offset_plus_length_blue) & fb->colors.bit_mask_blue);
                     return c |= ((c & ~((1 << fb->Var.blue.length) - 1)) >> fb->Var.blue.length);
                 }
        // ------------------------------------------------------------------- //
        case 24:
        case 32: return (rgb >> fb->Var.blue.offset) & fb->colors.bit_mask_blue;
    }
}

//############################################################################
int set_fb_cmap_to_was(struct frame_buffer_device* fb)
{
EZFB_DEBUGGING_FUNCTION_CALL_0
    if(fb->colors.was_grabbed)
    {
        memcpy((void*)fb->colors.red   , (void*)fb->colors.red_was   , 256);
        memcpy((void*)fb->colors.green , (void*)fb->colors.green_was , 256);
        memcpy((void*)fb->colors.blue  , (void*)fb->colors.blue_was  , 256);
        ret = set_fb_cmap(fb);
    }
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int set_fb_cmap_grays(struct frame_buffer_device* fb)
{
    int i;
EZFB_DEBUGGING_FUNCTION_CALL_0
    for(i = 0; i < (1 << fb->Var.bits_per_pixel); i++)
        set_fb_cmap_index(fb, i, (i % (1 << fb->colors.shortest_color)) << (NUM_BITS_IN_SHORT - fb->colors.shortest_color),
                                 (i % (1 << fb->colors.shortest_color)) << (NUM_BITS_IN_SHORT - fb->colors.shortest_color),
                                 (i % (1 << fb->colors.shortest_color)) << (NUM_BITS_IN_SHORT - fb->colors.shortest_color)
                         );
    if(fb->colors.was_grabbed)
    {
        ret = set_fb_cmap(fb);
        fb->colors.black = 0;
        fb->colors.white = (1 << fb->colors.shortest_color) - 1;
    }

EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int set_fb_cmap_reds(struct frame_buffer_device* fb)
{
    int i;
EZFB_DEBUGGING_FUNCTION_CALL_0
    for(i = 0; i < (1 << fb->Var.bits_per_pixel); i++)
        set_fb_cmap_index(fb, i, (i % (1 << fb->Var.red.length)) << (NUM_BITS_IN_SHORT - fb->Var.red.length), 0, 0);
    if(fb->colors.was_grabbed)
    {
        ret = set_fb_cmap(fb);
        fb->colors.black = 0;
        fb->colors.white = (1 << fb->Var.red.length) - 1;
    }

EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int set_fb_cmap_greens(struct frame_buffer_device* fb)
{
    int i;
EZFB_DEBUGGING_FUNCTION_CALL_0
    for(i = 0; i < (1 << fb->Var.bits_per_pixel); i++)
        set_fb_cmap_index(fb, i, 0, (i % (1 << fb->Var.green.length)) << (NUM_BITS_IN_SHORT - fb->Var.green.length), 0);
    if(fb->colors.was_grabbed)
    {
        ret = set_fb_cmap(fb);
        fb->colors.black = 0;
        fb->colors.white = (1 << fb->Var.green.length) - 1;
    }
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int set_fb_cmap_blues(struct frame_buffer_device* fb)
{
    int i;
EZFB_DEBUGGING_FUNCTION_CALL_0
    for(i = 0; i < (1 << fb->Var.bits_per_pixel); i++)
        set_fb_cmap_index(fb, i, 0, 0, (i % (1 << fb->Var.blue.length)) << (NUM_BITS_IN_SHORT - fb->Var.blue.length));
    if(fb->colors.was_grabbed)
    {
        ret = set_fb_cmap(fb);
        fb->colors.black = 0;
        fb->colors.white = (1 << fb->Var.blue.length) - 1;
    }
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int set_fb_cmap_utility(struct frame_buffer_device* fb)
{
EZFB_DEBUGGING_FUNCTION_CALL_0
    if(1 == fb->Var.bits_per_pixel)
    {
        set_fb_cmap_index(fb,  0, 0x0000, 0x0000, 0x0000);
        set_fb_cmap_index(fb,  1, 0xffff, 0xffff, 0xffff);
        fb->colors.black = 0;
        fb->colors.white = 1;
    }
    else if(2 == fb->Var.bits_per_pixel)
    {
        set_fb_cmap_index(fb,  0, 0x0000, 0x0000, 0x0000);
        set_fb_cmap_index(fb,  1, 0xffff, 0x0000, 0x0000);
        set_fb_cmap_index(fb,  2, 0x0000, 0xffff, 0xffff);
        set_fb_cmap_index(fb,  3, 0xffff, 0xffff, 0xffff);
        fb->colors.black = 0;
        fb->colors.white = 3;
    }
    else if((4 == fb->Var.bits_per_pixel) || (8 == fb->Var.bits_per_pixel))
    {   // default console colors
        set_fb_cmap_index(fb,  0, 0x0000, 0x0000, 0x0000);
        set_fb_cmap_index(fb,  1, 0x0000, 0x0000, 0xa000);
        set_fb_cmap_index(fb,  2, 0x0000, 0xa000, 0x0000);
        set_fb_cmap_index(fb,  3, 0x0000, 0xa000, 0xa000);
        set_fb_cmap_index(fb,  4, 0xa000, 0x0000, 0x0000);
        set_fb_cmap_index(fb,  5, 0xa000, 0x0000, 0xa000);
        set_fb_cmap_index(fb,  6, 0xa000, 0xa000, 0x0000);
        set_fb_cmap_index(fb,  7, 0xa000, 0xa000, 0xa000);
        set_fb_cmap_index(fb,  8, 0x4000, 0x4000, 0x4000);
        set_fb_cmap_index(fb,  9, 0x4000, 0x4000, 0xf000);
        set_fb_cmap_index(fb, 10, 0x4000, 0xf000, 0x4000);
        set_fb_cmap_index(fb, 11, 0x4000, 0xf000, 0xf000);
        set_fb_cmap_index(fb, 12, 0xf000, 0x4000, 0x4000);
        set_fb_cmap_index(fb, 13, 0xf000, 0x4000, 0xf000);
        set_fb_cmap_index(fb, 14, 0xf000, 0xf000, 0x4000);
        set_fb_cmap_index(fb, 15, 0xf000, 0xf000, 0xf000);
        fb->colors.black = 0;
        fb->colors.white = 15;
    }
    if(8 == fb->Var.bits_per_pixel)
    {
        int i;
        for(i = 0; i < 16; i++)
        {
            set_fb_cmap_index(fb, i +  16, i * 0x1000, i * 0x1000, i * 0x1000); // grays
            set_fb_cmap_index(fb, i +  32, i * 0x1000,          0,          0); // reds
            set_fb_cmap_index(fb, i +  48,          0, i * 0x1000,          0); // greens
            set_fb_cmap_index(fb, i +  64,          0,          0, i * 0x1000); // blues
            set_fb_cmap_index(fb, i +  80, i * 0x1000, i * 0x1000,          0); // yellows
            set_fb_cmap_index(fb, i +  96, i * 0x1000,          0, i * 0x1000); // magentas
            set_fb_cmap_index(fb, i + 112,          0, i * 0x1000, i * 0x1000); // cyans
            set_fb_cmap_index(fb, i + 128, i * 0x1000, i * 0x0800,          0); // oranges
            set_fb_cmap_index(fb, i + 144, i * 0x1000,          0, i * 0x0800); // hot pinks
            set_fb_cmap_index(fb, i + 160, i * 0x0800,          0, i * 0x1000); // purples
            set_fb_cmap_index(fb, i + 176,          0, i * 0x1000, i * 0x0800); // mints
            set_fb_cmap_index(fb, i + 192,          0, i * 0x0800, i * 0x1000); // light blues
            set_fb_cmap_index(fb, i + 208, i * 0x1000, i * 0x0800, i * 0x0800); // salmons
            set_fb_cmap_index(fb, i + 224, i * 0x0800, i * 0x0800, i * 0x1000); // light indigos
        }
        set_fb_cmap_index(fb, 240, 0xc000, 0xa000, 0x4000); // tan
        set_fb_cmap_index(fb, 241, 0x8000, 0x4000, 0x8000); // light purple
        set_fb_cmap_index(fb, 242, 0x8000, 0x8000, 0xffff); // light blue
        set_fb_cmap_index(fb, 243, 0xffff, 0x8000, 0x8000); // pink
        set_fb_cmap_index(fb, 244, 0xffff, 0x0000, 0x8000); // hot pink
        set_fb_cmap_index(fb, 245, 0x8000, 0x0000, 0x8000); // purple
        set_fb_cmap_index(fb, 246, 0xffff, 0x8000, 0x0000); // orange
        set_fb_cmap_index(fb, 247, 0x8000, 0x8000, 0x8000); // midtone gray
        set_fb_cmap_index(fb, 248, 0xffff, 0xffff, 0x0000); // real yellow
        set_fb_cmap_index(fb, 249, 0xffff, 0x0000, 0xffff); // real magenta
        set_fb_cmap_index(fb, 250, 0x0000, 0xffff, 0xffff); // real cyan
        set_fb_cmap_index(fb, 251, 0x0000, 0x0000, 0xffff); // real blue
        set_fb_cmap_index(fb, 252, 0x0000, 0xffff, 0x0000); // real green
        set_fb_cmap_index(fb, 253, 0xffff, 0x0000, 0x0000); // real red
        set_fb_cmap_index(fb, 254, 0x0000, 0x0000, 0x0000); // real black
        set_fb_cmap_index(fb, 255, 0xffff, 0xffff, 0xffff); // real white
        fb->colors.black = 0;
        fb->colors.white = 255;
    }
    if(fb->colors.was_grabbed)
        ret = set_fb_cmap(fb);
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int set_fb_cmap_smooth(struct frame_buffer_device* fb)
{
    int   i;
    float pi   = 4 * atan(1),
          Rfac = random_01() * (pi / (fb->colors.fbcmap.len / 4.0)),
          Roff = random_01() * 2 * pi,
          Gfac = random_01() * (pi / (fb->colors.fbcmap.len / 4.0)),
          Goff = random_01() * 2 * pi,
          Bfac = random_01() * (pi / (fb->colors.fbcmap.len / 4.0)),
          Boff = random_01() * 2 * pi;
EZFB_DEBUGGING_FUNCTION_CALL_0
    for(i = 0; i < (1 << fb->Var.bits_per_pixel); i++)
        set_fb_cmap_index(fb, i, (sin(Roff + Rfac * i) + 1) * 0x7fff,
                                 (sin(Goff + Gfac * i) + 1) * 0x7fff,
                                 (sin(Boff + Bfac * i) + 1) * 0x7fff    );
    if(fb->colors.was_grabbed)
    {
        ret =  set_fb_cmap(fb);
        ret &= find_black_and_white(fb);
    }
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int set_fb_cmap_random(struct frame_buffer_device* fb)
{
    int i;
EZFB_DEBUGGING_FUNCTION_CALL_0
    for(i = 0; i < 256; i++)
        set_fb_cmap_index(fb, i, (random() % 0x00010000),
                                 (random() % 0x00010000),
                                 (random() % 0x00010000) );
    if(fb->colors.was_grabbed)
    {
        ret =  set_fb_cmap(fb);
        ret &= find_black_and_white(fb);
    }
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
int set_fb_cmap_332(struct frame_buffer_device* fb)
{
    int i;
EZFB_DEBUGGING_FUNCTION_CALL_0
    switch(fb->Var.bits_per_pixel)
    {
        case  1: set_fb_cmap_index(fb, 0, 0x0000, 0x0000, 0x0000); // black
                 set_fb_cmap_index(fb, 1, 0xffff, 0xffff, 0xffff); // white
                 break;
        // ------------------------------------------------------------------- //
        case  2: set_fb_cmap_index(fb, 0, 0x0000, 0x0000, 0x0000); // black
                 set_fb_cmap_index(fb, 1, 0xffff, 0x0000, 0xffff); // magenta
                 set_fb_cmap_index(fb, 2, 0x0000, 0xffff, 0xffff); // cyan
                 set_fb_cmap_index(fb, 3, 0xffff, 0xffff, 0xffff); // white
                 break;
        // -------------------------------------------------------------------- //
        case  4: for(i = 0; i < 256; i += 17)
                     set_fb_cmap_index(fb, i / 17, (i & 0xe0) <<  8,
                                                   (i & 0x1c) << 11,
                                                   (i & 0x03) << 14  );
                 break;
        // ------------------------------------------------------------------- //
        case  8:
        case 15:
        case 16:
        case 24:
        case 32: for(i = 0; i < 256; i++)
                     set_fb_cmap_index(fb, i, (i & 0xe0) <<  8,
                                              (i & 0x1c) << 11,
                                              (i & 0x03) << 14  );
                 break;
        // ------------------------------------------------------------------- //
    }
    if(fb->colors.was_grabbed && (fb->Var.bits_per_pixel <= 8))
    {
        ret =  set_fb_cmap(fb);
        ret &= find_black_and_white(fb);
    }
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
void set_fb_cmap_index(struct frame_buffer_device* fb, int i, u_short r, u_short g, u_short b)
{
    if((i > 0) & (i < 256))
    {
        fb->colors.red   [i] = fb->Var.red.msb_right   ? (r >> (NUM_BITS_IN_SHORT - fb->Var.red.length   )) : r;
        fb->colors.green [i] = fb->Var.green.msb_right ? (g >> (NUM_BITS_IN_SHORT - fb->Var.green.length )) : g;
        fb->colors.blue  [i] = fb->Var.blue.msb_right  ? (b >> (NUM_BITS_IN_SHORT - fb->Var.blue.length  )) : b;
    }
}

//############################################################################
void get_fb_cmap_index(struct frame_buffer_device* fb, int i, u_short* r, u_short* g, u_short* b)
{
    if((i > 0) & (i < 256))
    {
        *r = fb->Var.red.msb_right   ? (fb->colors.red   [i] << (NUM_BITS_IN_SHORT - fb->Var.red.length   )) : fb->colors.red   [i];
        *g = fb->Var.green.msb_right ? (fb->colors.green [i] << (NUM_BITS_IN_SHORT - fb->Var.green.length )) : fb->colors.green [i];
        *b = fb->Var.blue.msb_right  ? (fb->colors.blue  [i] << (NUM_BITS_IN_SHORT - fb->Var.blue.length  )) : fb->colors.blue  [i];
    }
}

//############################################################################
u_short get_r_from_fb_cmap_index(struct frame_buffer_device* fb, int i)
{
    if((i > 0) & (i < 256))
        return (fb->Var.red.msb_right ? fb->colors.red[i] : fb->colors.red[i] >> NUM_BITS_IN_CHAR) & fb->colors.bit_mask_red;
    return 0;
}

//############################################################################
u_short get_g_from_fb_cmap_index(struct frame_buffer_device* fb, int i)
{
    if((i > 0) & (i < 256))
        return (fb->Var.green.msb_right ? fb->colors.green[i] : fb->colors.green[i] >> NUM_BITS_IN_CHAR) & fb->colors.bit_mask_green;
    return 0;
}

//############################################################################
u_short get_b_from_fb_cmap_index(struct frame_buffer_device* fb, int i)
{
    if((i > 0) & (i < 256))
        return (fb->Var.blue.msb_right ? fb->colors.blue[i] : fb->colors.blue[i] >> NUM_BITS_IN_CHAR) & fb->colors.bit_mask_blue;
    return 0;
}

//############################################################################
int cycle_fb_cmap(struct frame_buffer_device* fb, int from_index, int to_index, int direction)
{
    if(fb->colors.was_grabbed && direction && (from_index != to_index))
    {
        int i, j, cycle_span;

        if(from_index > to_index)
        {
            i          = from_index ;
            from_index = to_index   ;
            to_index   = i          ;
        }

        cycle_span = to_index - from_index;

        for(i = from_index; i <= to_index; i++)
        {
            j = (i + direction) % 256;
            fb->colors.red   [i] = fb->colors.red   [j];
            fb->colors.green [i] = fb->colors.green [j];
            fb->colors.blue  [i] = fb->colors.blue  [j];
        }

        return set_fb_cmap(fb);
    }
    return 0;
}

//############################################################################
//int fb_cmap_color_parade()
//{
//
//}

//############################################################################
int dump_ezfb_colors(struct frame_buffer_device* fb)
{
EZFB_DEBUGGING_FUNCTION_CALL
#if EZFB_API_MESSAGES
    printf("\n      EZFB COLORS:"                                                          );
    printf("\n      fb->colors.was_grabbed    = %u"  , fb->colors.was_grabbed              );
    printf("\n      fb->colors.white          = 0x%X", fb->colors.white                    );
    printf("\n      fb->colors.black          = 0x%X", fb->colors.black                    );
    printf("\n      fb->colors.bit_mask_red   = 0x%X", fb->colors.bit_mask_red             );
    printf("\n      fb->colors.bit_mask_green = 0x%X", fb->colors.bit_mask_green           );
    printf("\n      fb->colors.bit_mask_blue  = 0x%X", fb->colors.bit_mask_blue            );
    printf("\n      fb->colors.shortest_color = %u"  , fb->colors.shortest_color           );
    printf("\n      fb->colors.red            = 0x%X", (unsigned int) fb->colors.red       );
    printf("\n      fb->colors.green          = 0x%X", (unsigned int) fb->colors.green     );
    printf("\n      fb->colors.blue           = 0x%X", (unsigned int) fb->colors.blue      );
    printf("\n      fb->colors.red_was        = 0x%X", (unsigned int) fb->colors.red_was   );
    printf("\n      fb->colors.green_was      = 0x%X", (unsigned int) fb->colors.green_was );
    printf("\n      fb->colors.blue_was       = 0x%X", (unsigned int) fb->colors.blue_was  );
    printf("\n\n");fflush(stdout);
#endif
EZFB_DEBUGGING_FUNCTION_RETURN(fb->colors.was_grabbed)
};

//############################################################################
int dump_ezfb_cmap(struct frame_buffer_device* fb)
{
EZFB_DEBUGGING_FUNCTION_CALL
    if(fb->colors.was_grabbed)
    {
#if EZFB_API_MESSAGES
        printf("\n      FBCMAP:"                                                                );
        printf("\n      fbcmap.start             = %u"  , fb->colors.fbcmap.start               );
        printf("\n      fbcmap.len               = %u"  , fb->colors.fbcmap.len                 );
        printf("\n      fbcmap.red               = 0x%X", (unsigned int)fb->colors.fbcmap.red   );
        printf("\n      fbcmap.green             = 0x%X", (unsigned int)fb->colors.fbcmap.green );
        printf("\n      fbcmap.blue              = 0x%X", (unsigned int)fb->colors.fbcmap.blue  );
        printf("\n      fbcmap.transp            = 0x%X", (unsigned int)fb->colors.fbcmap.transp);
        printf("\n\n");fflush(stdout);
#endif
    }
    else
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "\nezfb ERROR: FB CMAP NOT GRABBED!");fflush(stderr);
#endif
    }
EZFB_DEBUGGING_FUNCTION_RETURN(fb->colors.was_grabbed)
}

//############################################################################
int dump_fb_cmap_values(struct frame_buffer_device* fb)
{
EZFB_DEBUGGING_FUNCTION_CALL
    if(fb->colors.was_grabbed)
    {
#if EZFB_API_MESSAGES
        int i;
        printf("\nFB CMAP SINCE THIS APPLICATION CHANGED IT:");
        printf("\n[index]. [ red ]. [green]. [blue ]");
        for(i = 0; i < 256; i++)
            printf("\n   %3d . 0x%4X . 0x%4X . 0x%4X", i, fb->colors.red   [i],
                                                          fb->colors.green [i],
                                                          fb->colors.blue  [i]  );
        printf("\n\n");fflush(stdout);
#endif
    }
    else
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "\nezfb ERROR: FB CMAP NOT GRABBED!");fflush(stderr);
#endif
    }
EZFB_DEBUGGING_FUNCTION_RETURN(fb->colors.was_grabbed)
}

//############################################################################
int dump_fb_cmap_was_values(struct frame_buffer_device* fb)
{
EZFB_DEBUGGING_FUNCTION_CALL
    if(fb->colors.was_grabbed)
    {
#if EZFB_API_MESSAGES
        int i;
        printf("\nFB CMAP BEFORE THIS APPLICATION CHANGED IT:");
        printf("\n[index]. [ red ]. [green]. [blue ]");
        for(i = 0; i < 256; i++)
            printf("\n   %3d . 0x%4X . 0x%4X . 0x%4X", i, fb->colors.red_was   [i],
                                                          fb->colors.green_was [i],
                                                          fb->colors.blue_was  [i]  );
        printf("\n\n");fflush(stdout);
#endif
    }
    else
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "\nezfb ERROR: FB CMAP NOT GRABBED!");fflush(stderr);
#endif
    }
EZFB_DEBUGGING_FUNCTION_RETURN(fb->colors.was_grabbed)
}

//############################################################################
int binfile_fb_cmap(struct frame_buffer_device* fb, char* file_name)
{
EZFB_DEBUGGING_FUNCTION_CALL_0
    if(fb->colors.was_grabbed)
    {
        int i, binfile_fd;
        char c;
        if(0 > (binfile_fd = open(file_name, (O_WRONLY | O_CREAT), (S_IRWXU | S_IRGRP | S_IROTH)))) // mode 744
        {
#if EZFB_API_ERRORS
            perror("\nezfb ERROR: opening output file failed");
#endif
            ret = 0;
        }
        else
        {
            ret = 1;
            for(i = 0; i < fb->colors.fbcmap.len; i++)
            {
                c = (char)get_r_from_fb_cmap_index(fb, i);
                write(binfile_fd, (void*)(&c), 1);
                c = (char)get_g_from_fb_cmap_index(fb, i);
                write(binfile_fd, (void*)(&c), 1);
                c = (char)get_b_from_fb_cmap_index(fb, i);
                write(binfile_fd, (void*)(&c), 1);
                c = 0;
                write(binfile_fd, (void*)(&c), 1);
            }
            close(binfile_fd);
        }
    }
    else
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "\nezfb ERROR: FB CMAP NOT GRABBED!");fflush(stderr);
#endif
    }
EZFB_DEBUGGING_FUNCTION_RETURN(ret)
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################

